----------------
-- File globals
----------------
local shortName = "SomeCooldowns"
local Addon = LibStub("AceAddon-3.0"):NewAddon(shortName, "AceConsole-3.0", "AceEvent-3.0")
Addon.name = "Some Cooldowns"
Addon.shortName = shortName
Addon.debug = true
Addon.settings = LibStub:GetLibrary(Addon.shortName .. "-Settings")
Addon.version = 0.2

local D = LibStub:GetLibrary("DoomCore-0.10")
local K = D.K
Addon.handler = D.Handler:New(Addon)

local inWorld
local frame
Addon.cooldowns = { spell = {}, item = {} }

local count = 0
--local showing = 0
local oldNumShown = {}
local loadDelay = 10
local sortDelay = 1
local abrupt = 3

--------
-- Core
--------

function Addon:OnInitialize()
  self.handler:Register({
    media = "LibSharedMedia-3.0",
    masque = "Masque",
    somebars = "SomeBars-Exports"
  }, nil, "Default")

  frame = CreateFrame("Frame", shortName, UIParent)
  frame.tex = frame:CreateTexture()
  frame.tex:SetAllPoints(frame)
  frame.tex:SetColorTexture(0, 1, 0, 0.5)
  frame.tex:Hide()

  frame:RegisterForDrag("LeftButton")
  frame:SetScript("OnDragStart", frame.StartMoving)
  frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
  frame:SetScript("OnReceiveDrag", function()
    local orig,_,dest,x,y = frame:GetPoint(1)
    local offset = (self.core.iconSize + self.core.padding * 2) / 2
    local a = D.growAnchors[orig]
    self.core.anchor = dest:lower()
    self.core.offsetX = math.floor(x*10)/10 + offset * a[3]
    self.core.offsetY =  math.floor(y*10)/10 + offset * a[4]
    self:Update()
  end)

  if self.lib.masque then
    frame.msq = self.lib.masque:Group(self.name)
  end

  self:RegisterEvent("PLAYER_ENTERING_WORLD", "OnLoad")
  self:RegisterEvent("GET_ITEM_INFO_RECEIVED", "OnItem")
  self:RegisterEvent("SPELL_UPDATE_COOLDOWN", "Update")
  self:RegisterEvent("SPELL_UPDATE_USABLE", "Update")
  self:RegisterEvent("SPELL_UPDATE_CHARGES", "Update")
  self:RegisterEvent("BAG_UPDATE", "Update")
  self:RegisterEvent("BAG_UPDATE_USABLE", "Update")
end

function Addon:OnLoad()
  inWorld = GetTime()
  self.core.group = self.core.group or {}
  for val in self:GetAll(self.core.group, "blacklist") do
    self.settings:Add("blacklist", val)
  end
  for val in self:GetAll(self.core.group, "whitelist") do
    self.settings:Add("whitelist", val)
  end
  self:SetFilter()
end

function Addon:OnItem()
  for valName, val in pairs(self.core.group or {}) do
    if type(valName) == "number" then
      local entry = self.settings.crawler:Get{ val.type, "args", tostring(valName) }
      if not entry and GetItemInfo(valName) then
        self:Add({ val.type }, valName)
      end
    end
  end
end

function Addon:Migrate()
  local oldVersion = self:GetVersion() or 0
  self:SetVersion(Addon.version)
  if oldVersion >= 0.2 then return end
  if self.core.rowGrowth == "up" then self.core.rowGrowth = "top" end
  if self.core.rowGrowth == "down" then self.core.rowGrowth = "bottom" end
end


function Addon:OnReset(registered)
  if registered == false and self:Migrate() then return end
end

function Addon:SetFilter(info, val)
  if not frame then return end
  local update = false
  if info then
    if self:ConfGet(info) ~= val then
      update = true
      self.cooldowns = { item = {}, spell = {} }
    end
    self:ConfSet(info, val)
  end
  for val in self:GetAll(self.core.group, "blacklist") do
    local cooldowns
    if type(val) == "number" then
      cooldowns = self.cooldowns.item
    else
      cooldowns = self.cooldowns.spell
    end
    cooldowns[val] = true
    update = true
  end

  self:Update(update)
end

function Addon:SetText(info, val)
  self:ConfSet(info, val)
  for _,v in ipairs({ frame:GetChildren() }) do
    v.cooldown:SetHideCountdownNumbers(not val)
  end
end

function Addon:SetReverse(info, val)
  self:ConfSet(info, val)
  for _,v in ipairs({ frame:GetChildren() }) do
    v.cooldown:SetReverse(val)
  end
end

function Addon:SetSwipeColor(info, r, g, b, a)
  self:ConfSet(info, r, g, b, a)
  for _,button in ipairs({ frame:GetChildren() }) do
    button.cooldown:SetSwipeColor(r, g, b, a)
  end
end


----------
-- Events
----------

function Addon:Sort()
  Addon:Output("Sorting")
  local children = { frame:GetChildren() }
  if #children == 0 then return end
  local remove = {}
  local add = {}
  local time = GetTime()
  --[[
  local suppressed,oldCount,oldTime = unpack(oldNumShown)
  oldCount = oldCount or 0
  self:Output(showing,time,"<-",suppressed,oldCount,math.floor(time - inWorld - loadDelay), oldTime and math.floor(time - oldTime))
  if suppressed then
    if showing < oldCount + abrupt then
      oldNumShown = { false, showing, time }
      frame:Show()
    end
  elseif showing > oldCount + abrupt and inWorld and time > inWorld + loadDelay then
    oldNumShown[1] = true
    frame:Hide()
  elseif not oldTime or time > oldTime + sortDelay then
    oldNumShown = { false, showing, time }
  end
  if showing == 0 then return end
  ]]--

  for _,button in ipairs(children) do
    if button:IsVisible() then
      local start,duration
      local type = button:GetAttribute("type")
      if type then
        local subject = button:GetAttribute(type)
        if subject then
          if type == "spell" then
            local charges, maxCharges, chargeStart, chargeDuration = GetSpellCharges(subject)
            if charges and charges < maxCharges and chargeStart and chargeStart > 0 and chargeDuration > 1.5 then
              start = chargeStart
              duration = chargeDuration * (maxCharges - charges)
            else
              start,duration = GetSpellCooldown(subject)
            end
          elseif type == "item" then
            start,duration = GetItemCooldown(subject)
          end
          if not start or not duration or duration <= 1.5 or start + duration + 1 <= time
              or not button.cooldown:IsVisible() or button.cooldown:GetCooldownDuration() == 0 then
            table.insert(remove, button)
          else
            table.insert(add, button)
          end
        end
      end
    end
  end

  for _,button in ipairs(remove) do
    button:Hide()
  end

  if #add == 0 then return end

  D.nilSort(add, function(e) return -e.expires end)

  D.makeGrid{
    frame        = frame,
    els          = add,
    parent       = UIParent,
    anchor       = Addon.core.anchor,
    x            = Addon.core.offsetX,
    y            = Addon.core.offsetY,
    size         = Addon.core.iconSize,
    padding      = Addon.core.padding,
    spacing      = Addon.core.spacing,
    grow         = Addon.core.grow,
    rowGrowth    = Addon.core.rowGrowth,
    columnGrowth = Addon.core.columnGrowth,
    limit        = Addon.core.limit,
    background   = Addon.core.background,
    border       = Addon.core.border,
    edge         = Addon.core.edge
  }

  if frame.msq then frame.msq:ReSkin() end
end

function Addon:OnSpellCooldown(name, icon)
  if not name then return end
  local record = 0
  local oldRecord, button
  local info = self.cooldowns.spell[name]
  if info then
    if info == true then return end
    oldRecord, button = unpack(info)
  else
    info = {}
    self.cooldowns.spell[name] = info
  end
  if oldRecord == true then return end
  local start, duration, cooldown
  local charges, maxCharges, chargeStart, chargeDuration = GetSpellCharges(name)
  if charges then
    record = charges
    if oldRecord == record then return end
    if not self.core.recharge and charges > 0 then return end
    if record == maxCharges then return end
    if oldRecord and button and record > oldRecord then
      ActionButton_ShowOverlayGlow(button)
      ActionButton_HideOverlayGlow(button)
    end
    if charges < maxCharges and chargeStart and chargeStart > 0 and chargeDuration > 1.5 then
      start = chargeStart
      duration = chargeDuration * (maxCharges - charges)
      cooldown = true
    end
  else
    if oldRecord == record then return end
    start, duration, cooldown = GetSpellCooldown(name)
  end
  if name and start and start > 0 and duration > 1.5 and cooldown then
    info[1] = record
    self:AddButton("spell", name, start, duration, icon)
    self:Output("Sorting with spell ",name)
    return true
  end
end
function Addon:OnItemCooldown(name)
  if not name then return end
  local oldRecord, button
  local info = self.cooldowns.item[name]
  if info then
    if info == true then return end
    oldRecord, button = unpack(info)
    if oldRecord then return end
  else
    info = {}
    self.cooldowns.item[name] = info
  end
  if oldRecord then return end
  local start, duration, cooldown = GetItemCooldown(name)
  if not start then return end

  if start > 0 and duration > 1.5 and cooldown then
    self.cooldowns.item[name][1] = 0
    self:AddButton("item", name, start, duration, GetItemIcon(name))
    self:Output("Sorting with item ",name)
    return true
  end
end

function Addon:Update(hard)
  local resort = false
  if C_LossOfControl.GetEventInfo(1) then return end
  if hard == true then
    self:Output("Hard reset")
    for _,button in ipairs({ frame:GetChildren() }) do
      button:Hide()
    end
  end
  for item in self:GetAll(self.core.group, "whitelist") do
    if type(item) == "number" then
      local fromCooldown = self:OnItemCooldown(item)
      if fromCooldown then
        self:Output("Sorted from whitelist item ",item)
      end
      resort = fromCooldown or resort
    else
      local icon = select(3, GetSpellInfo(item))
      local fromCooldown = self:OnSpellCooldown(item, icon)
      if fromCooldown then
        self:Output("Sorted from whitelist spell ",item)
      end
      resort = fromCooldown or resort
    end
  end

  if self.core.displaySpells then
    local index = 1
    while true do
      local _,id = GetSpellBookItemInfo(index, "spell")
      if not id then break end
      local spell,_,icon = GetSpellInfo(id)
      spell = GetSpellInfo(spell) or spell
      local fromCooldown = self:OnSpellCooldown(spell, icon)
      if fromCooldown then
        self:Output("Sorted from spell ",spell)
      end
      resort = fromCooldown or resort
      index = index + 1
    end
  end

  if self.core.displayItems then
    for slot=1,19 do
      local item = GetInventoryItemID("player", slot)
      local fromCooldown = self:OnItemCooldown(item)
      if fromCooldown then
        self:Output("Sorted from equipped item ",item)
      end
      resort = fromCooldown or resort
    end
    for bag = 0,NUM_BAG_SLOTS do
      for slot = 1,GetContainerNumSlots(bag) do
        local item = GetContainerItemID(bag, slot)
        local fromCooldown = self:OnItemCooldown(item)
        if fromCooldown then
          self:Output("Sorted from inventory item ",item)
        end
        resort = fromCooldown or resort
      end
    end
  end

  if self.core.displayToys then
    for i=1,C_ToyBox.GetNumToys() do
      local toy = C_ToyBox.GetToyFromIndex(i)
      if PlayerHasToy(toy) then
        local fromCooldown = self:OnItemCooldown(toy)
        if fromCooldown then
          self:Output("Sorted from toy ",toy)
        end
        resort = fromCooldown or resort
      end
    end
  end
  if resort then self:Sort() end
end

----------
-- Frames
----------
function Addon:UpdateFrame(info, r, g, b, a)
  if info then self:ConfSet(info, r, g, b, a) end
  local size = self.core.iconSize
  local lock = self.core.lock
  frame:SetMovable(not lock)
  frame:EnableMouse(not lock)
  if lock then
    frame.tex:Hide()
  else
    frame.tex:Show()
  end

  for _,button in ipairs({ frame:GetChildren() }) do
    button:SetSize(size, size)
  end

  self:Update()
  self:Sort()
end

function Addon:Add(info, val)
  local group
  if info[1] == "addToWhitelist" then
    group = "whitelist"
  elseif info[1] == "addToBlacklist" then
    group = "blacklist"
  else return end

  if not GetSpellInfo(val) then
    val = D.getItemID(val)
  end
  if not val then return end

  if info[1] == "addToBlacklist" then
    for _,b in ipairs({ frame:GetChildren() }) do
      local contained = b:GetAttribute(b:GetAttribute("type"))
      if contained == val then
        b:Hide()
        break
      end
    end
  end

  self:Make(self.core.group, group, val)

  self.settings:Add(group, val)

  self:Update()
end

function Addon:InitButton(button)
  button:Hide()
  ActionButton_ShowOverlayGlow(button)
  ActionButton_HideOverlayGlow(button)
  for _,child in ipairs{ button:GetChildren() } do
    if child.outerGlowOver then
      child.outerGlow:Hide()
      child.outerGlowOver:Hide()
      child.ants:Hide()
    end
  end
  button:SetAttribute("unit","player")
  button.cooldownFrame = _G[button:GetName() .. "Cooldown"]
  button.cooldownFrame:SetEdgeTexture("Interface\\Cooldown\\edge-LoC")
  button.cooldownFrame:SetScript("OnHide", self.Sort)
  button.cooldown:SetSwipeColor(self:ConfGet{ "color" })
  button.cooldown:SetHideCountdownNumbers(not self.core.text)
  button:Disable()
  button:SetMotionScriptsWhileDisabled(self.core.tooltip)
  button:SetScript("OnEnter", function(button)
    if not button:IsVisible() then return end
    if self.core.tooltipOverride then
      GameTooltip:SetOwner(button, "ANCHOR_" .. self.core.tooltipAnchor)
    else
      GameTooltip:SetOwner(button)
    end
    local buttonType = button:GetAttribute("type")
    if buttonType == "spell" then
      GameTooltip:SetSpellByID(select(7, GetSpellInfo(button:GetAttribute(buttonType))))
    elseif buttonType == "item" then
      GameTooltip:SetItemByID(button:GetAttribute(buttonType))
    end
    GameTooltip:Show()
  end)
  button:SetScript("OnLeave", function(button)
    if not button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
  end)
  button:SetScript("OnHide", function(button)
    --showing = showing - 1
    if button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
    local type = button:GetAttribute("type")
    if not type then return end
    local subject = button:GetAttribute(type)
    Addon:Output("Canceling ",type,subject)
    Addon.cooldowns[type][subject] = {}
  end)
end

function Addon:AddButton(buttonType, subject, time, duration, texture)
  for blacklisted in self:GetAll(self.core.group, "blacklist") do
    if subject == blacklisted
        or (buttonType == "spell" and GetSpellInfo(subject) == blacklisted)
        or (buttonType == "item" and GetItemInfo(subject) == GetItemInfo(blacklisted)) then
      return
    end
  end
  if self:ConfGetDebug{ "somebars" } then
    for frameName, alsoName, also in self:GetAll(self.lib.somebars, "also", 3) do
      if subject == alsoName
          or (buttonType == "spell" and GetSpellInfo(subject) == alsoName)
          or (buttonType == "item" and GetItemInfo(subject) == GetItemInfo(alsoName)) then
        return
      end
    end
  end
  local button
  local oldInfo = self.cooldowns[buttonType][subject]
  if oldInfo then
    button = oldInfo[2]
  else
    oldInfo = {}
    self.cooldowns[buttonType][subject] = oldInfo
  end
  if not button then
    for _,b in ipairs({ frame:GetChildren() }) do
      if not b:IsVisible() then
        button = b
        break
      end
    end
  end
  if not button then
    count = count + 1
    button = CreateFrame("Button", frame:GetName() .. count, frame, "ActionButtonTemplate")
    self:InitButton(button)
    if frame.msq then frame.msq:AddButton(button) end
  end
  self.cooldowns[buttonType][subject][2] = button
  button.icon:SetTexture(texture)
  local iconSize = self.core.iconSize
  button:SetSize(iconSize, iconSize)
  button:SetAttribute("type", buttonType)
  button:SetAttribute(buttonType, subject)
  button:GetNormalTexture():Hide()
  button.cooldownFrame:SetCooldown(time, duration)
  button.cooldownFrame:SetReverse(self.core.reverse)
  button.expires = time + duration
  button:Show()
end

function Addon:UpdateTooltips(info, val)
  self:ConfSet(info, val)
  local tooltip = self.core.tooltip
  local override = self.core.tooltipOverride
  local anchor = self.core.tooltipAnchor
  for _,button in ipairs({ frame:GetChildren() }) do
    if button:IsMouseOver() then
      if not tooltip then button:GetScript("OnLeave")(button)
      elseif override then
        GameTooltip:SetOwner(button,anchor)
      else
        GameTooltip:SetOwner(button)
      end
    end
    button:SetMotionScriptsWhileDisabled(tooltip)
  end
end